@@ -1,3 +1,4 @@
+from deprecated import deprecated
 import json
 import logging
 import time
@@ -10,9 +11,11 @@ import aiohttp
 
 from eyepop.endpoint import Endpoint
 from eyepop.exceptions import PopNotStartedException, PopConfigurationException, PopNotReachableException
-from eyepop.worker.worker_jobs import WorkerJob, _UploadJob, _LoadFromJob, _UploadStreamJob
+from eyepop.worker.worker_client_session import WorkerClientSession
+from eyepop.worker.worker_jobs import WorkerJob, _UploadFileJob, _LoadFromJob, _UploadStreamJob, _LoadFromAssetUuidJob
 from eyepop.worker.load_balancer import EndpointLoadBalancer
 from eyepop.worker.worker_syncify import SyncWorkerJob
+from eyepop.worker.worker_types import Pop, VideoMode, ComponentParams
 
 log = logging.getLogger('eyepop')
 log_requests = logging.getLogger('eyepop.requests')
@@ -23,14 +26,18 @@ MAX_RETRY_TIME_SECS = 30.0
 FORCE_REFRESH_CONFIG_SECS = (61.0 * 61.0)
 
 
-class WorkerEndpoint(Endpoint):
+class WorkerEndpoint(Endpoint, WorkerClientSession):
     """
     Endpoint to an EyePop.ai worker.
     """
 
-    def __init__(self, secret_key: str, eyepop_url: str, pop_id: str, auto_start: bool, stop_jobs: bool,
+    def __init__(self, secret_key: str | None, access_token: str | None,
+                 eyepop_url: str, pop_id: str, auto_start: bool, stop_jobs: bool,
                  job_queue_length: int, is_sandbox: bool, request_tracer_max_buffer: int):
-        super().__init__(secret_key, eyepop_url, job_queue_length, request_tracer_max_buffer)
+        super().__init__(
+            secret_key=secret_key, access_token=access_token, eyepop_url=eyepop_url,
+            job_queue_length=job_queue_length, request_tracer_max_buffer=request_tracer_max_buffer
+        )
         self.pop_id = pop_id
         self.auto_start = auto_start
         self.stop_jobs = stop_jobs
@@ -45,8 +52,14 @@ class WorkerEndpoint(Endpoint):
         self.last_fetch_config_error = None
         self.last_fetch_config_error_time = None
 
+        # old way, TODO remove
         self.pop_comp = None
+        self.model_refs = []
         self.post_transform = None
+
+        # new way
+        self.pop = None
+
         self.add_retry_handler(404, self._retry_404)
 
     async def _retry_404(self, status_code: int, failed_attempts: int) -> bool:
@@ -57,17 +70,40 @@ class WorkerEndpoint(Endpoint):
             self.worker_config = None
             return True
 
-    async def _disconnect(self):
+    async def _disconnect(self, timeout: float | None = None):
+        client_timeout = None
+        if timeout is not None:
+            client_timeout = aiohttp.ClientTimeout(total=timeout)
+        if (self.is_dev_mode and self.pop_id == 'transient'
+                and self.worker_config is not None and self.worker_config['pipeline_id'] is not None):
+            try:
+                base_url = await self.dev_mode_base_url()
+                delete_pipeline_url = f'{base_url}/pipelines/{self.worker_config["pipeline_id"]}'
+                headers = {}
+                authorization_header = await self._authorization_header()
+                if authorization_header is not None:
+                    headers['Authorization'] = authorization_header
+                log_requests.debug('before DELETE %s', delete_pipeline_url)
+                await self.client_session.delete(delete_pipeline_url, headers=headers, timeout=client_timeout)
+                log_requests.debug('after DELETE %s', delete_pipeline_url)
+            except Exception as e:
+                log.exception(e, exc_info=True)
+            finally:
+                del self.worker_config["pipeline_id"]
+
         if self.sandbox_id is not None:
             try:
                 base_url = await self.dev_mode_base_url()
                 delete_sandbox_url = f'{base_url}/sandboxes/{self.sandbox_id}'
-                headers = {'Authorization': await self._authorization_header()}
+                headers = {}
+                authorization_header = await self._authorization_header()
+                if authorization_header is not None:
+                    headers['Authorization'] = authorization_header
                 log_requests.debug('before DELETE %s', delete_sandbox_url)
-                await self.client_session.delete(delete_sandbox_url, headers=headers)
+                await self.client_session.delete(delete_sandbox_url, headers=headers, timeout=client_timeout)
                 log_requests.debug('after DELETE %s', delete_sandbox_url)
             except Exception as e:
-                log.exception("error at disconnect", e)
+                log.exception(e, exc_info=True)
             finally:
                 self.sandbox_id = None
 
@@ -85,7 +121,10 @@ class WorkerEndpoint(Endpoint):
             config_url = f'{self.eyepop_url}/workers/config'
         else:
             config_url = f'{self.eyepop_url}/pops/{self.pop_id}/config?auto_start={self.auto_start}'
-        headers = {'Authorization': await self._authorization_header()}
+        headers = {}
+        authorization_header = await self._authorization_header()
+        if authorization_header is not None:
+            headers['Authorization'] = authorization_header
         try:
             log_requests.debug('before GET %s', config_url)
             async with self.client_session.get(config_url, headers=headers) as response:
@@ -102,7 +141,10 @@ class WorkerEndpoint(Endpoint):
                 log_requests.debug('after GET %s: 401, about to retry with fresh access token', config_url)
                 self.token = None
                 self.expire_token_time = None
-                headers = {'Authorization': await self._authorization_header()}
+                headers = {}
+                authorization_header = await self._authorization_header()
+                if authorization_header is not None:
+                    headers['Authorization'] = authorization_header
                 async with self.client_session.get(config_url, headers=headers) as retried_response:
                     self.worker_config = await retried_response.json()
         except aiohttp.ClientConnectionError as e:
@@ -119,7 +161,10 @@ class WorkerEndpoint(Endpoint):
 
         if self.is_sandbox and self.sandbox_id is None:
             create_sandbox_url = f'{base_url}/sandboxes'
-            headers = {'Authorization': await self._authorization_header()}
+            headers = {}
+            authorization_header = await self._authorization_header()
+            if authorization_header is not None:
+                headers['Authorization'] = authorization_header
             log_requests.debug('before POST %s', create_sandbox_url)
             async with self.client_session.post(create_sandbox_url, headers=headers) as response:
                 response_json = await response.json()
@@ -127,17 +172,24 @@ class WorkerEndpoint(Endpoint):
             self.sandbox_id = response_json
 
         if self.pop_id == 'transient':
-            self.pop_comp = 'identity'
             if self.sandbox_id is None:
                 start_pipeline_url = f'{base_url}/pipelines'
             else:
                 start_pipeline_url = f'{base_url}/pipelines?sandboxId={self.sandbox_id}'
+            if self.pop is not None:
+                body = {"pop": self.pop, "source": {"sourceType": "NONE"},
+                        "idleTimeoutSeconds": 60, "logging": ["out_meta"], "videoOutput": "no_output"}
+            else:
+                if self.pop_comp is None:
+                    self.pop_comp = 'identity'
+                body = {'inferPipelineDef': {'pipeline': self.pop_comp, 'modelRefs': self.model_refs},
+                        'postTransformDef': {'transform': self.post_transform}, "source": {"sourceType": "NONE"},
+                        "idleTimeoutSeconds": 60, "logging": ["out_meta"], "videoOutput": "no_output"}
 
-            body = {'inferPipelineDef': {'pipeline': self.pop_comp},
-                    'postTransformDef': {'transform': self.post_transform}, "source": {"sourceType": "NONE", },
-                    "idleTimeoutSeconds": 60, "logging": ["out_meta"], "videoOutput": "no_output"}
-
-            headers = {'Authorization': await self._authorization_header()}
+            headers = {}
+            authorization_header = await self._authorization_header()
+            if authorization_header is not None:
+                headers['Authorization'] = authorization_header
             log_requests.debug('before POST %s', start_pipeline_url)
             async with self.client_session.post(start_pipeline_url, headers=headers, json=body) as response:
                 response_json = await response.json()
@@ -150,7 +202,10 @@ class WorkerEndpoint(Endpoint):
         if self.is_dev_mode and self.stop_jobs:
             stop_jobs_url = f'{await self.dev_mode_pipeline_base_url()}/source?mode=preempt&processing=sync'
             body = {'sourceType': 'NONE'}
-            headers = {'Authorization': await self._authorization_header()}
+            headers = {}
+            authorization_header = await self._authorization_header()
+            if authorization_header is not None:
+                headers['Authorization'] = authorization_header
             log_requests.debug('before PATCH %s', stop_jobs_url)
             async with self.client_session.patch(stop_jobs_url, headers=headers, json=body) as response:
                 pass
@@ -159,7 +214,10 @@ class WorkerEndpoint(Endpoint):
         if self.is_dev_mode and self.pop_comp is None:
             # get current pipeline string and store
             get_url = await self.dev_mode_pipeline_base_url()
-            headers = {'Authorization': await self._authorization_header()}
+            headers = {}
+            authorization_header = await self._authorization_header()
+            if authorization_header is not None:
+                headers['Authorization'] = authorization_header
             log_requests.debug('before GET %s', get_url)
             async with self.client_session.get(get_url, headers=headers) as response:
                 response_json = await response.json()
@@ -175,20 +233,49 @@ class WorkerEndpoint(Endpoint):
         else:
             self.load_balancer = EndpointLoadBalancer(self.worker_config['endpoints'])
 
+    async def session(self) -> dict:
+        session = await super().session()
+        session['popId'] = self.pop_id
+        if self.worker_config:
+            session['baseUrl'] = self.worker_config['base_url']
+            session['pipelineId'] = self.worker_config['pipeline_id']
+        session['sandboxId'] = self.sandbox_id
+        return session
+
+    async def get_pop(self) -> Pop | None:
+        return self.pop
+
+    async def set_pop(self, pop: Pop):
+        if not self.is_dev_mode:
+            raise PopConfigurationException(self.pop_id, 'set_pop_comp not supported in production mode')
+        response = await self.pipeline_patch('pop', content_type='application/json',
+                                             data=pop.model_dump_json())
+        self.pop = pop
+        self.pop_comp = None
+        self.model_refs = []
+        self.post_transform = None
+
+        return response
+
+    @deprecated(version='1.0.0', reason='Use get_pop() instead')
     async def get_pop_comp(self) -> str:
         return self.pop_comp
 
-    async def set_pop_comp(self, pop_comp: str = None):
+    @deprecated(version='1.0.0', reason='Use set_pop() instead')
+    async def set_pop_comp(self, pop_comp: str = None, model_refs: list[dict] = []):
         if not self.is_dev_mode:
             raise PopConfigurationException(self.pop_id, 'set_pop_comp not supported in production mode')
         response = await self.pipeline_patch('inferencePipeline', content_type='application/json',
-                                             data=json.dumps({'pipeline': pop_comp}))
+                                             data=json.dumps({'pipeline': pop_comp, 'modelRefs': model_refs}))
         self.pop_comp = pop_comp
+        self.model_refs = model_refs
         return response
 
+    @deprecated(version='1.0.0', reason='Use get_pop() instead')
     async def get_post_transform(self) -> str:
         return self.post_transform
 
+    @deprecated(version='1.0.0', reason='Use set_pop() instead')
     async def set_post_transform(self, transform: str = None):
         if not self.is_dev_mode:
             raise PopConfigurationException(self.pop_id, 'set_post_transform not supported in production mode')
@@ -197,10 +284,6 @@ class WorkerEndpoint(Endpoint):
         self.post_transform = transform
         return response
 
-    '''
-    Deprecated
-    '''
-
     async def list_models(self) -> list[dict]:
         warnings.warn("list_models for development use only", DeprecationWarning)
         if self.sandbox_id is None:
@@ -265,24 +348,72 @@ class WorkerEndpoint(Endpoint):
         else:
             pass
 
-    async def upload(self, location: str, params: dict | None = None,
-                     on_ready: Callable[[WorkerJob], None] | None = None) -> WorkerJob | SyncWorkerJob:
-        job = _UploadJob(location=location, params=params, session=self, on_ready=on_ready,
-                         callback=self.metrics_collector)
+    async def upload(
+            self,
+            location: str,
+            video_mode: VideoMode | None = None,
+            params: list[ComponentParams] | None = None,
+            on_ready: Callable[[WorkerJob], None] | None = None
+    ) -> WorkerJob | SyncWorkerJob:
+        job = _UploadFileJob(
+            location=location,
+            video_mode=video_mode,
+            component_params=params,
+            session=self, on_ready=on_ready,
+            callback=self.metrics_collector
+        )
         await  self._task_start(job.execute())
         return job
 
-    async def upload_stream(self, stream: BinaryIO, mime_type: str, params: dict | None = None,
-                            on_ready: Callable[[WorkerJob], None] | None = None) -> WorkerJob | SyncWorkerJob:
-        job = _UploadStreamJob(stream, mime_type, params=params, session=self, on_ready=on_ready,
-                               callback=self.metrics_collector)
+    async def upload_stream(
+            self,
+            stream: BinaryIO,
+            mime_type: str,
+            video_mode: VideoMode | None = None,
+            params: list[ComponentParams] | None = None,
+            on_ready: Callable[[WorkerJob], None] | None = None
+    ) -> WorkerJob | SyncWorkerJob:
+        job = _UploadStreamJob(
+            stream=stream,
+            mime_type=mime_type,
+            video_mode=video_mode,
+            component_params=params,
+            session=self,
+            on_ready=on_ready,
+            callback=self.metrics_collector
+        )
         await self._task_start(job.execute())
         return job
 
-    async def load_from(self, location: str, params: dict | None = None,
-                        on_ready: Callable[[WorkerJob], None] | None = None) -> WorkerJob | SyncWorkerJob:
-        job = _LoadFromJob(location=location, params=params, session=self, on_ready=on_ready,
-                           callback=self.metrics_collector)
+    async def load_from(
+            self,
+            location: str,
+            params: list[ComponentParams] | None = None,
+            on_ready: Callable[[WorkerJob], None] | None = None
+    ) -> WorkerJob | SyncWorkerJob:
+        job = _LoadFromJob(
+            location=location,
+            component_params=params,
+            session=self,
+            on_ready=on_ready,
+            callback=self.metrics_collector
+        )
+        await self._task_start(job.execute())
+        return job
+
+    async def load_asset(
+            self,
+            asset_uuid: str,
+            params: list[ComponentParams] | None = None,
+            on_ready: Callable[[WorkerJob], None] | None = None
+    ) -> WorkerJob | SyncWorkerJob:
+        job = _LoadFromAssetUuidJob(
+            asset_uuid=asset_uuid,
+            component_params=params,
+            session=self,
+            on_ready=on_ready,
+            callback=self.metrics_collector
+        )
         await self._task_start(job.execute())
         return job
 
@@ -330,6 +461,7 @@ class WorkerEndpoint(Endpoint):
         if self.last_fetch_config_success_time is not None and self.last_fetch_config_success_time < time.time() - FORCE_REFRESH_CONFIG_SECS:
             self.worker_config is None
 
+        failed_attempts = 0
         retried_re_auth = False
         retried_re_config = False
         start_time = time.time()
@@ -350,7 +482,10 @@ class WorkerEndpoint(Endpoint):
 
             url = f'{entry.base_url}/pipelines/{entry.pipeline_id}/{url_path_and_query}'
 
-            headers = {'Authorization': await self._authorization_header()}
+            headers = {}
+            authorization_header = await self._authorization_header()
+            if authorization_header is not None:
+                headers['Authorization'] = authorization_header
             if accept is not None:
                 headers['Accept'] = accept
             if content_type is not None:
@@ -374,18 +509,18 @@ class WorkerEndpoint(Endpoint):
 
                 return response
             except aiohttp.ClientResponseError as e:
-                if not retried_re_auth and e.status == 401:
-                    # auth token might have just expired
-                    log_requests.debug('after %s %s: 401, about to retry with fresh access token', method, url)
-                    self.token = None
-                    self.expire_token_time = None
-                    retried_re_auth = True
-                elif e.status == 404:
+                if e.status == 404:
+                    # in load balanced configuration, we overwrite the standard 404 handler
                     entry.mark_error()
                     log_requests.debug('after %s %s: 404, about to retry fail-over', method, url)
                 else:
-                    log_requests.exception('unexpected error', e)
-                    raise e
+                    failed_attempts += 1
+                    if e.status not in self.retry_handlers:
+                        log_requests.exception('unexpected error', e)
+                        raise e
+                    if not await self.retry_handlers[e.status](e.status, failed_attempts):
+                        log_requests.exception('unexpected error', e)
+                        raise e
             except aiohttp.ClientConnectionError:
                 entry.mark_error()
                 log_requests.debug('after %s %s: 404, about to retry fail-over', method, url)
